r""" Parametric-AAA

"""

from itertools import product
import numpy as np
from copy import deepcopy as copy
import scipy.linalg
from iterprinter import IterationPrinter

from .rational import RationalBarycentric

def _build_cauchy(x,y):
	with np.errstate(divide = 'ignore', invalid = 'ignore'):
		C = 1./(np.tile(x.reshape(-1,1), (1,len(y))) - np.tile(y.reshape(1,-1), (len(x),1)))
		# If we have a divide by zero error, we reason that for that row,
		# only the matching column is non-zero
		for row in np.argwhere(~np.all(np.isfinite(C), axis = 1)):
			C[row] = np.zeros(C.shape[1])
			C[row, np.argmin(np.abs(x[row] - y)).flatten()] = 1
	return C 

def eval_paaa(Xe, X, y, I, b, basis, order):
	d = X.shape[1]

	C = np.ones((len(Xe),1), dtype = X.dtype)
	for i in range(d):
		Ct = _build_cauchy(Xe[:, i], basis[i])
		C = np.vstack([C[:,j] * Ct[:,k] for j, k in product(range(C.shape[1]), range(Ct.shape[1]))]).T

	num = np.einsum('ij,j...->i...', C, [bi*yi for bi, yi in zip(b,y[order])] )
	denom = np.einsum('ij,j...->i...', C, b)
	
	with np.errstate(divide = 'ignore', invalid = 'ignore'):
		return np.einsum('i...,i->i...', num, 1./denom)

def paaa(X, y, verbose = True, maxiter = 100, tol = None):
	r"""

	"""
	X = np.array(X)
	y = np.array(y)
	assert X.shape[0] == y.shape[0], "X and y must have the same first dimension"

	if tol is None:
		tol = 1e-10

	d = X.shape[1]
	r = np.zeros(y.shape)

	degree = np.zeros(d, dtype = np.int)


	if verbose:
		printer = IterationPrinter(it = '4d', degree = f'{2+3*d:d}s', norm = '20.10e', cond = '10.2e')
		printer.print_header(it = 'iter', degree = 'degree', norm = 'norm mismatch', cond = 'cond #')
		printer.print_iter(norm = np.linalg.norm(y))
	# indices of points where we interpolate
	I = np.zeros(y.shape[0], dtype = np.bool)

	
	mismatch = np.copy(y)
	best = None
	for it in range(maxiter):
		# compute the pointwise-residual
		residual = np.sum(np.abs(mismatch), axis = tuple(range(1,len(y[0].shape)+1)))
		# zero out points we've already identified so we don't re-evaluate
		residual[I] = 0
		
		# find the point with the largest residual and add that to the interpolation set
		i = np.argmax(residual)
		I[i] = True
		
		# determine points in tensor-product barycentric basis
		basis = [np.array(list(set(X[I,i]))).reshape(-1) for i in range(d)]
		degree = np.array([len(b)-1 for b in basis], dtype = np.int)

		# With this algorithm we need to "complete the square"
		# by adding all terms of the form
		I = np.all( [np.any([X[:,i] == bi for bi in basis[i]], axis = 0) for i in range(d)], axis = 0)
		#print("I", np.argwhere(I).flatten())
		assert np.prod([len(b) for b in basis]) == np.sum(I), "Missing points in tensor product grid"
		
		# Build the multivariate Cauchy matrix
		C = np.ones((np.sum(~I),1), dtype = X.dtype)
		for i in range(d):
			Ct = _build_cauchy(X[~I, i], basis[i])
			C = np.vstack([C[:,j] * Ct[:,k] for j, k in product(range(C.shape[1]), range(Ct.shape[1]))]).T

		# Build the Loewner matrix associated with each input
		Lten = []
		# The issue is we need to match the ordering of I to that in C
		order = [ int(np.argwhere( (X == xx).all(axis=1))) for xx in product(*basis)]
		for idx in np.ndindex(y[0].shape):
			Lten.append( (C.T * y[(~I,*idx)]).T - C*y[(order,*idx)] )
		L = np.vstack(Lten)
	
		# Compute coefficients for denominator polynomial
		U, s, VH = scipy.linalg.svd(L, full_matrices = False, compute_uv = True, overwrite_a = True)
		b = VH.conj().T[:,-1] 
		
		if len(s) >= 2:
			with np.errstate(divide='ignore',invalid='ignore'):
				cond = s[0]*np.sqrt(2)/(s[-2] - s[-1])
		else:
			cond = None
	
		r = eval_paaa(X, X, y, I, b, basis, order)
		if np.all(np.isfinite(r)):
			best = (np.copy(I), np.copy(b), copy(basis), copy(order))
			assert np.all(np.isclose(r[I], y[I])), "Constructed rational approximant failed to correctly interpolate"
		else:
			if verbose:
				print("Encountered nan in evaluation")
			break
			
		mismatch = y - r

		mismatch_norm = np.linalg.norm(mismatch)

		if verbose:
			degree_str = '(' + ','.join([f'{di:2d}' for di in degree]) + ')'
			printer.print_iter(it = it, norm = mismatch_norm, degree = degree_str, cond = cond) 

		if mismatch_norm < tol:
			if verbose: print('terminated due to small mismatch')
			break
	
	return best


class ParametricAAARationalApproximation(RationalBarycentric):
	def __init__(self, tol = None, verbose = True, maxiter = 100):
		self.tol = None
		self.verbose = verbose
		self.maxiter = maxiter
		self.degree = None

	def fit(self, X, y):
		self.X = np.copy(X)
		self.y = np.copy(y)
		self._I, self._b, self._basis, self._order = paaa(X, y, verbose = self.verbose, maxiter = self.maxiter)
		self.degree = np.array([len(bi)-1 for bi in self._basis], dtype = np.int)	


	def __call__(self, X):
		return eval_paaa(X, self.X, self.y, self._I, self._b, self._basis, self._order)

	@property
	def interpolation_points(self):
		return np.array(list(product(*self._basis)))
		



import warnings
import numpy as np
from scipy.linalg import eig, eigvals, hessenberg
from .basis import PolynomialBasis
from .polynomial import Polynomial


def lagrange_roots(nodes, weights, coef, deflation = True):
	r""" Compute the roots of a Lagrange polynomial

	This implements the deflation algorithm from LC14



	Parameters
	----------
	nodes: numpy.array (n,)
		Nodes :math:`x_j` where the polynomial value is defined.
	weights: numpy.array (n,)
		The barycentric weights :math:`prod_{k\ne j} (x_j - x_k)^{-1}`
	coef: numpy.array (n,)
		The coeffients :math:`c_j` defining the value of the polynomial at each point;
		i.e., :math:`p(x_j) = c_j`.
	"""
	n = len(nodes)
	assert (n == len(weights)) and  (n == len(coef)), "Dimensions of nodes, weights, and coef should be the same"

	
	# Build the RHS of the generalized eigenvalue problem
	C0 = np.zeros((n+1, n+1), dtype=np.complex)
	C0[1:n+1,1:n+1] = np.diag(nodes)

	# LHS for generalized eigenvalue problem	
	C1 = np.eye(n+1, dtype=np.complex)
	C1[0, 0] = 0

	# scaling
	coef = coef / np.linalg.norm(coef)
	weights = weights / np.linalg.norm(weights)

	C0[0,1:n+1] = coef
	C0[1:n+1,0] = weights


	# balancing [LC14, eq. 29]
	coef0 = np.copy(coef)
	weights0 = np.copy(weights)
	s = np.array([1.]+[np.sqrt(np.abs(wj/aj)) if np.abs(aj) > 0 else 1 for (wj, aj) in zip(weights, coef)])
	C0 = np.diag(1/s) @ C0 @ np.diag(s)

	# Apply a rotation to make the first weight real
	angle = np.angle(C0[1,0])
	if np.isfinite(angle):
		C0[1:,0] *= np.exp(-1j*angle)
	else:
		print("Rotation failed", angle)
		deflation = False

	if deflation:
		#C0[1,0] must be real for Householder to reflect correctly
		assert np.abs(C0[1,0].imag) < 1e-10, "C0[1,0]: %g + I %g" % (C0[1,0].real, C0[1,0].imag)
		#Householder Reflector
		u = np.copy(C0[1:,0]) # = w scaled
		u[0] += np.linalg.norm(C0[1:,0]) # (w) scaled
		H = np.eye(n, dtype=complex) - 2 * np.outer(u,u.conjugate())/(np.linalg.norm(u)**2)
		G2 = np.zeros((n+1, n+1), dtype=complex)
		G2[0,0] = 1
		G2[1:,1:] = H
		C0 = G2 @ C0 @ G2
		C1 = G2 @ C1 @ G2
		H1, P1 = hessenberg(C0[1:,1:], calc_q=True, overwrite_a = False)
		G3 = np.zeros((n+1, n+1), dtype=complex)
		G3[0,0] = 1
		G3[1:,1:] = P1.T.conjugate()
		G4 = np.eye(n+1, dtype=complex)
		G4[0:2,0:2] = [[0,1],[1,0]]
		H1 = G4.dot(G3.dot(C0.dot(G3.T.conjugate())))[1:,1:]
		B1 = G4.dot(G3.dot(C1.dot(G3.T.conjugate())))[1:,1:]

		# Givens Rotation
		G5 = np.eye(n, dtype=complex)
		a = H1[0,0]
		b = H1[1,0]
		c = a / np.sqrt(a**2 + b**2)
		s = b / np.sqrt(a**2 + b**2)
		G5[0:2,0:2] = [[c.conjugate(), s.conjugate()],[-s,c]]

		H2 = G5.dot(H1)[1:,1:]
		B2 = G5.dot(B1)[1:,1:]
		try:
			return eigvals(H2, B2)
		except np.linalg.linalg.LinAlgError as e:
			raise e 
	else:
		# Compute the eigenvalues
		# As this eigenvalue problem has a double root at infinity, we ignore the division by zero warning
		with warnings.catch_warnings():
			warnings.filterwarnings("ignore", message='divide by zero encountered in true_divide',
									category=RuntimeWarning)
			ew = eigvals(C0, C1, overwrite_a=False)
		ew = ew[np.isfinite(ew).flatten()]
		assert len(ew) == len(coef) - 1, "Error: too many infinite eigenvalues encountered"

		return ew



def lagrange_vandermonde(nodes, weights, X):
	r""" Build the Vandermonde matrix associated with 
	"""
	x = X.flatten()
	assert len(x) == len(X), "Input must be one dimensional"

	with np.errstate(divide = 'ignore', invalid = 'ignore'):
		# The columns of the Vandermonde matrix 
		V = np.array([w/(x - n) for n, w in zip(nodes, weights)]).T
		for row in np.argwhere(~np.all(np.isfinite(V), axis = 1)):
			V[row] = 0
			V[row, np.argmin(np.abs(x[row] - nodes)).flatten()] = 1.
		
		denom = np.sum(V, axis = 1)
		V /= denom[:, None]
		
	
	return V


class LagrangePolynomialBasis(PolynomialBasis):
	r""" Constructs a Lagrange polynomial basis in barycentric form

	See: BT04
	"""
	def __init__(self, nodes):
		self.nodes = np.array(nodes).flatten()
		assert len(nodes) == len(self.nodes), "Input must be one-dimensional"

		# BT 04, eq. (3.2)
		# w[j] = prod_{j\ne k} 1./(node[j] - node[k])
		self.weights = 1./np.array([
			np.prod(self.nodes[k] - self.nodes[0:k]) * np.prod(self.nodes[k] - self.nodes[k+1:]) for k in range(len(self.nodes))
			])

	def vandermonde(self, X):
		return lagrange_vandermonde(self.nodes, self.weights, X)
	
	def roots(self, coef, deflation = True):
		r"""
		"""
		return lagrange_roots(self.nodes, self.weights, coef, deflation = deflation)


class LagrangePolynomialInterpolant(Polynomial):
	def __init__(self, X, y):
		self.basis = LagrangePolynomialBasis(X)
		self.coef = np.copy(y)
	
	@property
	def nodes(self):
		return self.basis.nodes





import numpy as np
from copy import deepcopy as copy
from .aaa import _build_cauchy
from .arnoldi import ArnoldiPolynomialBasis
from .skiter import linearized_ratfit
from .polynomial import Polynomial
from .lagrange import LagrangePolynomialInterpolant
from .basis import MonomialPolynomialBasis, LegendrePolynomialBasis
from .rational import RationalFunction, RationalRatio
from iterprinter import IterationPrinter
import scipy.linalg


def _solve_linearized_vecfit(num_basis, denom_basis, y):
	A = np.hstack([ num_basis, -(denom_basis[:,1:].T * y).T ])
	b = y
	x, res, rank, s = scipy.linalg.lstsq(A, b, overwrite_a = True, overwrite_b = True)
	a = x[0:num_basis.shape[1]]
	b = np.hstack([1, x[num_basis.shape[1]:]])
	return a, b, s[0]/s[-1]

# TODO: Implement a DGB15a variant of the linear solve that allows array-valued y

# TODO: Add matrix-weight option

def vecfit(X, y, num_degree, denom_degree, verbose = True, 
	Basis = ArnoldiPolynomialBasis, poles0 = 'linearized',
	maxiter = 50, ftol = 1e-10, btol = 1e-7):
	r"""Implements Vector Fitting 
	
	See: GS99

	Parameters
	----------

	poles0: ['GS', 'linearized', array-like]
		Specifies how the initial poles are to be selected
		* 'GS': as recommeneded by Gustavsen & Semlyen, linearized spaced between largest imaginary value
		* 'linearized', perform a linearized rational fitting
		* array-like: specify an array of denom_degree initial poles
	"""
	assert num_degree >= 0 and denom_degree >= 0, "numerator and denominator degrees must be nonnegative integers"
	assert num_degree + 1 >= denom_degree, "Vector fitting requires denominator degree to be at most one less than numerator degree"


	if verbose:
		printer = IterationPrinter(it = '4d', res = '20.10e', delta = '10.4e', bnorm = '10.4e', cond = '10.4e')
		printer.print_header(it = 'iter', res = 'residual norm', delta = 'Δ fit', bnorm = '‖b - e₀‖₂', cond = 'condion #') 

	if isinstance(poles0, str):
		if poles0 == 'linearized':
			# Generate initial poles by one step of SK iteration (i.e., the linearized ratfit)
			numerator, denominator = linearized_ratfit(X, y, num_degree, denom_degree, simultaneous = True)
			poles = denominator.roots().flatten()
		elif poles0 == 'GS':
			# Generate initial poles as recommened in GS99, Sec. 3.2 (eqns. 9-10)
			im_max = np.max(np.abs(X.imag))
			assert im_max > 0
			poles = -im_max/100 + 1j*np.linspace(-im_max, im_max, denom_degree)
		else:
			raise NotImplementedError 
	else:
		assert len(poles0) == denom_degree, "Number of poles must match the degree of the denominator"
		poles = np.array(poles0)	

	# Construct the Vandermonde matrix for the remaining terms
	if num_degree - denom_degree >= 0:
		bonus_basis = Basis(X, num_degree - denom_degree)
		V = bonus_basis.basis()
	else:
		bonus_basis = None
		V = np.zeros((len(y),0))

	r_old = np.zeros(y.shape)

	best_fit = {'residual_norm':np.inf}

	for it in range(maxiter):
		C = _build_cauchy(X, poles)

		num_basis = np.hstack([C, V])
		denom_basis = np.hstack([np.ones((len(X), 1)), C])

		a, b, cond = _solve_linearized_vecfit(num_basis, denom_basis, y)
		b_norm = b[0] #np.linalg.norm(b)
		b /= b_norm
		a /= b_norm
	
		# Compute the rational approximation
		r = (num_basis @ a) / (denom_basis @ b)

		residual_norm = np.linalg.norm( (y - r).flatten(), 2)
		delta_norm = np.linalg.norm( (r_old - r).flatten(), 2)
		b_norm = np.linalg.norm(b[1:]/b[0], np.inf)

		if residual_norm < best_fit['residual_norm']:
			best_fit['residual_norm'] = residual_norm
			best_fit['a'] = a
			best_fit['b'] = b
			best_fit['poles'] = poles

		if verbose:
			printer.print_iter(it = it, res = residual_norm, delta = delta_norm, bnorm = b_norm, cond = cond)	

		if it == maxiter - 1:
			break

		if delta_norm < ftol:
			if verbose: print("terminated due to small change in approximation")
			break

		if b_norm < btol:
			if verbose: print("terminated due to denominator being approximately constant")
			break
	
		# Update roots only if we are going to continue the iteration
		# (if we update these we change the polynomial values)
		# This is the root finding approach that Gustavsen takes in Vector Fitting
		# See Gus06: eq. 5
		poles = np.linalg.eigvals(np.diag(poles) - np.outer(np.ones(len(poles)), b[1:]))

		r_old = r		

	a = best_fit['a']
	b = best_fit['b']
	poles = best_fit['poles']

	if bonus_basis:
		bonus_poly = Polynomial(bonus_basis, a[len(poles):])
	else:
		bonus_poly = Polynomial(MonomialPolynomialBasis(X, 0), 0*a[0:1])

	return a[:len(poles)], b, poles, bonus_poly	


class VectorFittingRationalFunction(RationalRatio):
	def __init__(self, a, b, poles, bonus_poly = None):
		self._a = np.copy(a)
		self._b = np.copy(b)
		self.poles = np.copy(poles)
	
		self.bonus_poly = copy(bonus_poly)

	def eval(self, X):
		C = _build_cauchy(X, self.poles)
		num = C @ self._a
		denom = self._b[0] + C @ self._b[1:]
		if self.bonus_poly is not None:
			num += self.bonus_poly(X)

		return num/denom

			

class VectorFittingRationalApproximation(VectorFittingRationalFunction):
	def __init__(self, num_degree, denom_degree, *args, **kwargs):
		self.num_degree = int(num_degree)
		self.denom_degree = int(denom_degree)
		self.args = args
		self.kwargs = kwargs
	
	def fit(self, X, y):
		self._a, self._b, self.poles, self.bonus_poly = vecfit(X, y, self.num_degree, self.denom_degree, *self.args, **self.kwargs)

	



from .version import __version__

# Utilities
from .sorted_norm import *

# Polynomial utils
from .index import *
from .basis import *
from .polynomial import *
from .lagrange import *
from .arnoldi import *

# Rational 
from .rational import *
from .vecfit import *
from .aaa import *
from .paaa import *
from .skiter import *



r""" Vandermonde with Arnoldi basis


"""
import numpy as np
from .basis import PolynomialBasis
from .index import *

def _update_rule(indices, idx):
	diff = indices - idx
	j = np.min(np.argwhere( (np.sum(np.abs(diff), axis = 1) <= 1) & (np.min(diff, axis = 1) == -1)))
	i = int(np.argwhere(diff[j] == -1))
	return i, j	

#def _update_rule_max(idx, ids):
#	# Determine which column to multiply by
#	diff = idx - ids
#	# Here we pick the *least* recent column that is one off
#	j = np.min(np.argwhere( (np.sum(np.abs(diff), axis = 1) <= 1) & (np.min(diff, axis = 1) == -1)))
#	i = int(np.argwhere(diff[j] == -1))
#	return i, j	


def vandermonde_arnoldi_CGS(X, degree, weight = None, mode = None):
	r""" Multivariate Vandermode with Arnoldi using classical Gram-Schmidt with reorthogonalization

	Notes
	-----
	* The use of Classical Gram-Schmidt with reorthogonalization
	  was suggested by Yuji Nakatsukasa to improve performance.
	  This uses matrix operations rather than the vector operations
	  of modified Gram-Schmidt allowing the use of more efficient 
	  BLAS3 operations.


	Parameters
	----------
	X: np.array (M, dim)
		Input coordinates
	degree: int or list of ints
		Polynomial degree.  If an int, a total degree polynomial is constructed;
		if a list, the list must be length m and a maximum degree polynomial is
		constructed.
	weight: None or np.array (M,)
		Initial vector in the Arnoldi iteration
	mode: None or ['total', 'max']
		What type of polynomial basis to construct; only matters if dim>1.
		If None, the type of basis will be automatically detected. 
	"""
	M, dim = X.shape

	if weight is None:
		weight = np.ones(M)

	if mode is None:
		try:
			degree = int(degree)
			mode = 'total'
		except (TypeError, ValueError):
			mode = 'max'
			degree = np.copy(degree)

	if mode == 'total':
		indices = total_degree_index(dim, degree)
	elif mode == 'max':
		indices = max_degree_index(degree)
	
	Q = np.zeros((M, len(indices)), dtype = X.dtype)
	R = np.zeros((len(indices), len(indices)), dtype = X.dtype)

	iter_indices = enumerate(indices)	

	# In the first iteration we simply orthogonalize 
	k, idx = next(iter_indices)
	q = np.array(weight, dtype = X.dtype)
	R[0,0] = np.linalg.norm(q)
	Q[:,0] = q/R[0,0]


	for k, idx in iter_indices:
		i, j = _update_rule(indices, idx)
		# Form new column	
		q = X[:,i] * Q[:,j]
		
		# see Alg. 6.1 in Bjo94
		# rather than doing the explicit check for orthogonality,
		# we just go ahead and do two iterations as this is sufficient
		# for double precision accuracy 
		for it in range(2):
			s = Q[:,:k].conj().T @ q
			q -= Q[:,:k] @ s
			R[:k,k] +=s

		R[k,k] = np.linalg.norm(q)
		Q[:,k] = q/R[k,k]

	return Q, R, indices



def vandermonde_arnoldi_eval(X, R, indices, mode, weight = None):
	r"""
	"""

	X = np.array(X)
	M, m = X.shape
	W = np.zeros((X.shape[0], len(indices)), dtype = X.dtype)
	if weight is None:
		weight = np.ones(M, dtype = X.dtype)	
	

	iter_indices = enumerate(indices)

	# First column
	next(iter_indices)
	W[:,0] = weight/R[0,0]
	
	# Now work on the remaining columns
	for k, ids in iter_indices:
		i, j = _update_rule(indices[:k], ids)
		# Form new column
		w = X[:,i] * W[:,j]

		# Perform orthogonalizations
		w -= W[:,:k] @ R[:k,k]
		
		W[:,k] = w/R[k,k]

	return W


def vandermonde_arnoldi_eval_der(X, R, indices, mode, weight = None, V = None):
	if V is None:
		V = vandermonde_arnoldi_eval(X, R, indices, mode, weight = weight)
		
	M = X.shape[0]
	N = R.shape[1]
	n = X.shape[1]
	DV = np.zeros((M, N, n), dtype = (R[0,0] * X[0,0]).dtype)

	for ell in range(n):
		index_iterator = enumerate(indices)
		next(index_iterator)
		for k, ids in index_iterator:
			i, j = _update_rule(indices[:k], ids)
			# Q[:,k] = X[:,i] * Q[:,j] - sum_s Q[:,s] * R[s, k]
			if i == ell:
				DV[:,k,ell] = V[:,j] + X[:,i] * DV[:,j,ell] - DV[:,0:k,ell] @ R[0:k,k] 
			else:
				DV[:,k,ell] = X[:,i] * DV[:,j,ell] - DV[:,0:k,ell] @ R[0:k,k] 
			DV[:,k,ell] /= R[k,k]	
	
	return DV


class ArnoldiPolynomialBasis(PolynomialBasis):
	r""" A polynomial basis constructed using Vandermonde with Arnoldi

	Parameters
	----------
	X: array-like (M,m)
		Input coordinates
	degree: int or list of ints
		Polynomial degree.  If an int, a total degree polynomial is constructed;
		if a list, the list must be length m and a maximum degree polynomial is
		constructed.
	"""
	def __init__(self, X, degree, weight = None):
		self.X = np.copy(np.atleast_2d(X))
		self.dim = self.X.shape[1]
		try:
			self.degree = int(degree)
			self.mode = 'total'
		except (TypeError, ValueError):
			self.degree = np.copy(degree).astype(np.int)
			self.mode = 'max'
	
		self._Q, self._R, self._indices = vandermonde_arnoldi_CGS(self.X, self.degree, weight = weight)
		
	def basis(self):
		return self._Q   

	def vandermonde(self, X, weight = None):
		return vandermonde_arnoldi_eval(X, self._R, self._indices, self.mode, weight = weight)

	def vandermonde_derivative(self, X, weight = None):
		if np.array_equal(X, self.X):
			return vandermonde_arnoldi_eval_der(X, self._R, self._indices, self.mode, weight = weight, V = self._Q)
		else:
			return vandermonde_arnoldi_eval_der(X, self._R, self._indices, self.mode, weight = weight)

	def roots(self, coef, *args, **kwargs):
		from .basis import LegendrePolynomialBasis
		from .polynomial import PolynomialApproximation
		y = self.basis() @ coef
		poly = PolynomialApproximation(self.degree, Basis = LegendrePolynomialBasis)
		poly.fit(self.X, y)
		roots = poly.roots()
		
		return roots




import numpy as np
from itertools import product


def fixed_degree_index(dim, degree):
	r""" Generate all combinations where sum == degree
	"""
	if dim == 1:
		return np.array([[degree]]).astype(np.int)
	else:
		idx = fixed_degree_index(dim - 1, degree)
		idx = np.hstack([np.zeros((idx.shape[0],1), dtype = np.int), idx])
		indices = [idx]
		for i in range(1, degree+1):
			idx = fixed_degree_index(dim - 1, degree - i) 
			idx = np.hstack([i*np.ones((idx.shape[0],1), dtype = np.int), idx])
			indices.append(idx)
		return np.vstack(indices)

def total_degree_index(dim, degree):
	r"""

	Parameters
	----------
	dim: int, positive
		Number of dimensions
	degree: int, nonnegative
		Total degree of the polynomial

	Returns
	-------
	indices: np.array (M, dim)
		indices of polynomials 
	"""

	assert dim > 0, "Dimension must be positive"
	assert degree >=0, "Degree must be nonnegative"
		
	if dim == 1:
		return np.arange(degree+1).reshape(-1,1)	

	indices = [fixed_degree_index(dim, k) for k in range(degree+1)]
	return np.vstack(indices)	


def max_degree_index(degree):
	r""" List the indices of a maximum degree polynomial



	Parameters
	----------
	degree: array-like
		

	Returns
	-------
	index: list of tuples
		List of indices ordered as needed for 
	"""

	return np.array([ids for ids in product(*[range(d+1) for d in degree])])







""" Sequential linear programing approach for minimax optimization

"""

import numpy as np
from iterprinter import IterationPrinter
from scipy.optimize import OptimizeResult, approx_fprime, linprog, Bounds
from scipy.optimize._constraints import new_bounds_to_old
import cvxopt as co

def slp(fun, x0, args = (), jac = None, bounds = None, constraints = (), hess = None, hessp = None, verbose = True,
	maxiter = 100, xtol = 1e-6, c_armijo = 0.01, bt_maxiter = 30, **options):
	r""" Sequential linear programming approach for constrained optimization problems

	Solves 
	
	min_x  f(x)
	s.t.   lbc <= c(x) <= ubc 
           lb <= x <= ub

	"""

	nfev = 0
	njev = 0

	if jac is None or jac is False:
		assert False

	# Normalize constraints format as a list
	if not isinstance(constraints, list):
		constraints = [constraints]

	if verbose:
		iterprinter = IterationPrinter(it = '4d', obj = '20.14e', alpha = '8.3e', delta = '8.3e')
		iterprinter.print_header(it = 'iter', obj = "objective fun.", alpha = 'step length', delta = 'trust region radius')

	x = np.copy(x0)
	delta = 1e0

	stop = False
	it = 0

	fx = fun(x)


	if verbose:
		iterprinter.print_iter(it = it, obj = fx)

	while it < maxiter and not stop:
		it += 1
		# Generate linearized constraints
		Acon = []
		bcon = []
		for con in constraints:
			fx = con.fun(x)
			Jx = con.jac(x)

			# fx + Jx @ p <= ub
			I = np.isfinite(con.ub)
			Acon.append(Jx[I])
			bcon.append(con.ub[I] - fx[I])
			# lb <= fx + Jx @ p
			I = np.isfinite(con.lb)
			Acon.append(-Jx[I])
			bcon.append(fx[I] - con.lb[I])	


		fx = fun(x)
		# Generate search direction
		g = jac(x)	
		
		# Convert the bounds into a format for linprog
		if isinstance(bounds, Bounds):
			#linprog_bounds = [(lbi, ubi) for lbi, ubi in zip(bounds.lb, bounds.ub)]
			linprog_bounds = new_bounds_to_old(np.maximum(bounds.lb - x, -delta), np.minimum(bounds.ub - x, delta), len(x0)) 
		elif bounds is None:
			# Since Linprog assumes non-negative, we have to specify infinite lower bounds 
			linprog_bounds = [(None, None) for i in enumerate(x0)]	
		else:
			raise NotImplementedError
			#linprog_bounds = bounds

		# Solve LP
		if False:
			res = linprog(g, A_ub = np.vstack(Acon), b_ub = np.hstack(bcon), bounds = linprog_bounds)
			if not res.success:
				if verbose: print("linear program failed with message", res.message)
				return OptimizeResult(
						x = x, 
						success = False,
						status = -1,
						message = 'LP solve failed with message ' + res.message,
						fun = fx,
						jac = g, 	
						nit = it,
						nfev = nfev,
						njev = njev,)

			p = res.x
		else:
			co.solvers.options['show_progress'] = False
			G = co.matrix(np.vstack(Acon + [np.eye(len(x)), -np.eye(len(x))] ))
			h = co.matrix(np.hstack(bcon + [b[1] for b in linprog_bounds] + [-b[0] for b in linprog_bounds])) 
			sol = co.solvers.lp(co.matrix(g),G, h)
			#print(sol)
			p = np.array(sol['x']).flatten()

	
		# Backtracking line search	
		pred_decrease = g.T @ p
		for alpha in [0.5**k for k in range(bt_maxiter)]:
			xn = x + alpha*p
			fxn = fun(xn)
			# Check if satisfy Armijo conditions
			improved = (fxn	<= fx + c_armijo*alpha*pred_decrease)	
			feasible = True
			for con in constraints:
				cxn = con.fun(xn)
				feasible = feasible & (np.all(cxn <= con.ub) and np.all(con.lb <= cxn))
			
			if feasible and improved:
				break

		if not np.isclose(alpha,1):
			delta = np.max(np.abs(alpha*p))
		else:
			delta *=2
		

		if not (feasible and improved):
			return OptimizeResult(
					x = x, 
					success = False,
					status = -2,
					message = 'unable to find a successful step',
					fun = fx,
					jac = g, 	
					nit = it,
					nfev = nfev,
					njev = njev,)
				
		if verbose:
			iterprinter.print_iter(it=it, obj = fxn, alpha =alpha, delta = delta)

		if np.max(np.abs(alpha*p)) < xtol:
			if verbose: print("small change in parameters")
			return OptimizeResult(
					x = x, 
					success = True,
					status = 0,
					message = 'small change in objective',
					fun = fx,
					jac = g, 	
					nit = it,
					nfev = nfev,
					njev = njev,)
		x = xn
			
	return OptimizeResult(
			x = x, 
			success = True,
			status = 0,
			message = 'maximum number of iterations exceeded',
			fun = fx,
			jac = g, 	
			nit = it,
			nfev = nfev,
			njev = njev,)


def sl1lp(fun, x0, args = (), jac = None, bounds = None, constraints = (), verbose = True,
	maxiter = 100, **options):
	r"""

	Solve a constrained optimization problem in the l1 relaxation (see, Nocedal and Wright Sec. 18.5)

	""" 

	if jac is None or jac is False:
		assert False

	# Convert the bounds into a format for linprog
	if isinstance(bounds, Bounds):
		#linprog_bounds = [(lbi, ubi) for lbi, ubi in zip(bounds.lb, bounds.ub)]
		linprog_bounds = new_bounds_to_old(bounds.lb, bounds.ub, len(x0)) 
	elif bounds is None:
		# Since Linprog assumes non-negative, we have to specify infinite lower bounds 
		linprog_bounds = [(None, None) for i in enumerate(x0)]	
	else:
		linprog_bounds = bounds



	# Normalize constraints format as a list
	if not isinstance(constraints, list):
		constraints = [constraints]

	if verbose:
		printer = IterationPrinter(it = '4d', obj = '20.8e',  con= '20.8e', alpha = '8.3e')
		printer.print_header(it = 'iter', obj = "objective fun.", con = 'constraint violation',  alpha = 'step length')


	x = np.copy(x0)
	fx = fun(x)
	con_violation = 0
	for con in constraints:
		cx = con.fun(x)
		con_violation += np.sum(np.maximum(con.lb - cx, 0))
		con_violation += np.sum(np.maximum(cx - con.ub, 0))


	if verbose:
		printer.print_iter(it = 0, obj = fx, con = con_violation)
	mu = 1
	it = 0
	stop = False

	while it <= maxiter and not stop:
		it += 1
		
		assert False



import numpy as np
import scipy.linalg
import scipy.optimize
from .basis import *
from .polynomial import *
from .skiter import *
from .rational_ratio import *
from copy import deepcopy


class RationalFunction:
	def __call__(self, X):
		return self.eval(X)
	
	def eval(self, X):
		return self.__call__(X)

class RationalApproximation(RationalFunction):
	def __init__(self, num_degree, denom_degree):
		self.num_degree = num_degree
		self.denom_degree = denom_degree


class RationalRatio(RationalFunction):
	def __init__(self, numerator, denominator):
		self.numerator = deepcopy(numerator)
		self.denominator = deepcopy(denominator)

	@property
	def P(self):
		return self.numerator.basis.basis()

	@property
	def Q(self):
		return self.denominator.basis.basis()

	@property
	def a(self):
		return self.numerator.coef

	@property
	def b(self):
		return self.denominator.coef
	
	def eval(self, X):
		p = self.numerator(X)
		q = self.denominator(X)
		return p/q	



	def refine(self, X, y, norm = 2, verbose = False, **kwargs):
		a, b = rational_ratio_optimize(y, self.P, self.Q, self.a, self.b, norm = norm, verbose = verbose, **kwargs)

		self.numerator.coef = a
		self.denominator.coef = b	

		#if verbose:
		#	res_norm = np.linalg.norm( (self.P @ a)/(self.Q @ b) - y, norm)
		#	print(f"final residual norm {res_norm:21.15e}")


class LinearizedRationalApproximation(RationalApproximation, RationalRatio):
	def __init__(self, num_degree, denom_degree, **kwargs):
		RationalApproximation.__init__(self, num_degree, denom_degree)
		self.kwargs = kwargs

	def fit(self, X, y):
		self.numerator, self.denominator = linearized_ratfit(X, y, self.num_degree, self.denom_degree, **self.kwargs)

class SKRationalApproximation(RationalApproximation, RationalRatio):
	r"""

	Parameters
	----------
	
	"""

	def __init__(self, num_degree, denom_degree, refine = False, norm = 2, 
		Basis = None, rebase = True, maxiter = 20, verbose = True, xtol = 1e-7):

		RationalApproximation.__init__(self, num_degree, denom_degree)
		self._refine = refine
		self.norm = norm
		self.xtol = float(xtol)
		#if self.norm != 2:
		#	raise NotImplementedError

		self.maxiter = int(maxiter)
		self.verbose = verbose
		self.rebase = rebase
		if Basis is None:
			Basis = LegendrePolynomialBasis
		
		self.Basis = Basis

		self.numerator = None
		self.denominator = None

	def fit(self, X, y, denom0 = None):

		X = np.array(X)
		y = np.array(y)
		assert X.shape[0] == y.shape[0], "X and y do not have the same number of rows"

		if self.rebase:
			self.numerator, self.denominator, self.hist = skfit_rebase(
				X, y, self.num_degree, self.denom_degree,
				maxiter = self.maxiter, verbose = self.verbose, norm = self.norm,
				history = True, xtol = self.xtol, denom0 = denom0,
				)
		else:
			num_basis = self.Basis(X, self.num_degree)	
			denom_basis = self.Basis(X, self.denom_degree)	
			P = num_basis.basis()
			Q = denom_basis.basis()
		
			a, b, self.hist = skfit(y, P, Q, maxiter = self.maxiter, verbose = self.verbose, norm = self.norm, history = True, 
				xtol = self.xtol, denom0 = denom0)

			self.numerator = Polynomial(num_basis, a)
			self.denominator = Polynomial(denom_basis, b)
		
		if self._refine:
			self.refine(X, y, norm = self.norm)




class RationalBarycentric(RationalFunction):
	r"""
	"""
	def __init__(self, degree):
		self.degree = int(degree)
		assert self.degree >= 0, "Degree must be non-negative"

	@property
	def num_degree(self):
		return self.degree

	@property
	def denom_degree(self):
		return self.degree






			



r""" Tools to find a locally optimal rational approximation in ratio format

In the ratio format, we assume we have bases P and Q for the numerator and denominator 
"""

import numpy as np
import scipy.optimize
from scipy.optimize import LinearConstraint, NonlinearConstraint, Bounds
from .slp import sl1lp, slp

################################################################################
# Residual and jacobian  
################################################################################

def _rational_residual_real(x, P, Q, y):
	m = P.shape[1]
	n = Q.shape[1]
	Pa = P @ x[:m]
	Qb = Q @ x[-n:]
	return Pa/Qb - y 

def _rational_jacobian_real(x, P, Q):
	m = P.shape[1]
	n = Q.shape[1]
	Pa = P @ x[:m]
	Qb = Q @ x[-n:]
	return np.hstack([			
			np.multiply((1./Qb)[:,None], P),				
			np.multiply(-(Pa/Qb**2)[:,None], Q),
		])
			

def _rational_residual_complex(x, P, Q, y):
	m = P.shape[1]
	n = Q.shape[1]
	a = x[:2*m].view(complex)
	b = x[-2*n:].view(complex)
	res = (P @ a)/(Q @ b) - y
	return res.view(float)


def _rational_jacobian_complex(x, P, Q):
	m = P.shape[1]
	n = Q.shape[1]
	
	a = x[:2*m].view(complex)
	b = x[-2*n:].view(complex)
	Pa = P @ a
	Qb = Q @ b
	J = np.hstack([			
			np.multiply((1./Qb)[:,None], P),				
			np.multiply(-(Pa/Qb**2)[:,None], Q),
		])
	JRI = np.zeros((J.shape[0]*2, J.shape[1]*2), dtype = np.float)
	JRI[0::2,0::2] = J.real
	JRI[1::2,1::2] = J.real
	JRI[0::2,1::2] = -J.imag
	JRI[1::2,0::2] = J.imag
	return JRI


################################################################################
# 2-norm optimization proceedures
################################################################################

def _rational_ratio_optimize_pin_2norm_real(y, P, Q, a0, b0, **kwargs):
	m, n = P.shape[1], Q.shape[1]
	k = np.argmax(np.abs(b0))
	x0 = np.hstack([a0, b0[:k], b0[k+1:]])
	xfun = lambda x: np.hstack([x[:m+k], b0[k], x[m+k:]])
	
	mask = np.ones(m+n, dtype = np.bool)
	mask[m+k] = 0
	res = lambda x: _rational_residual_real(xfun(x), P, Q, y)
	jac = lambda x: _rational_jacobian_real(xfun(x), P, Q)[:,mask]

	result = scipy.optimize.least_squares(res, x0, jac, **kwargs)
	x = xfun(result.x)
	return x[:m], x[m:]


def _rational_ratio_optimize_pin_2norm_complex(y, P, Q, a0, b0, **kwargs):
	m, n = P.shape[1], Q.shape[1]
	k = np.argmax(np.abs(b0))
	x0 = np.hstack([a0, b0[:k], b0[k+1:]])
	xfun = lambda x: np.hstack([x[:2*(m+k)], b0[k].real, b0[k].imag, x[2*(m+k):]])
	
	mask = np.ones(2*(m+n), dtype = np.bool)
	mask[2*(m+k):2*(m+k)+2] = 0
	res = lambda x: _rational_residual_complex(xfun(x), P, Q, y)
	jac = lambda x: _rational_jacobian_complex(xfun(x), P, Q)[:,mask]

	result = scipy.optimize.least_squares(res, x0.view(float), jac, **kwargs)
	x = xfun(result.x)
	return x[:2*m].view(complex), x[2*m:].view(complex)
	

################################################################################
# sup-norm optimization utilities
################################################################################

# setup the objective for the constraint
def _rational_residual_squared_abs_complex(x, P, Q, y):
	r = _rational_residual_complex(x, P, Q, y)
	return r[::2]**2 + r[1::2]**2

def _rational_jacobian_squared_abs_complex(x, P, Q, y):
	r = _rational_residual_complex(x, P, Q, y)
	J = _rational_jacobian_complex(x, P, Q)
	Jt = np.zeros((r.shape[0]//2, x.shape[0]))
	Jt += 2*np.diag(r[0::2]) @ J[0::2,:]
	Jt += 2*np.diag(r[1::2]) @ J[1::2,:]
	return Jt	

def _rational_ratio_inf_complex(y, P, Q, a0, b0):
	r"""

	Solving as 

	min_{x, t} 0.5 * t**2
	st.        Re[f_j(x)]^2 + Im[f_j(x)]^2 - t^2 <= 0

	"""

	# Introduce a slack variable t to represent the maximum
	fun = lambda xt: 0.5*xt[-1]**2
	def grad(xt):
		g = np.zeros(xt.shape)
		g[-1] = xt[-1]
		return g

	def hess(xt):
		H = np.zeros( (len(xt), len(xt)) )
		H[-1,-1] = 1
		return H

	def con(xt):
		x = xt[:-1]
		t = xt[-1]
		r =  _rational_residual_squared_abs_complex(x, P, Q, y) - t**2
		return r

	def con_jac(xt):
		x = xt[:-1]
		t = xt[-1]
		J = _rational_jacobian_squared_abs_complex(x, P, Q, y)
		Jt = np.hstack([J, -2*t*np.ones((J.shape[0], 1)) ])
		return Jt

	lb = -np.inf*np.ones(P.shape[0])
	ub = np.zeros(P.shape[0])
	constraint = NonlinearConstraint(con, lb, ub, jac = con_jac)


	t0 = np.max(np.abs( (P @ a0)/(Q @ b0) - y))
	xt0 = np.hstack([a0.view(float), b0.view(float), t0])
	
	lb_var = -np.inf*np.ones(xt0.shape)
	lb_var[-1] = 0
	ub_var = np.inf*np.ones(xt0.shape)
	#print("t0", t0)

	# TODO: Replace with a call to SLP
	res = scipy.optimize.minimize(fun, xt0, jac = grad, constraints = constraint, method = slp,
		bounds = Bounds(lb_var, ub_var), options = dict(maxiter = 50))
	xt = res.x
	x = xt[:-1]
	a = x[:2*P.shape[1]].view(complex)
	b = x[-2*Q.shape[1]:].view(complex)
	#print("t", xt[-1])
	#print("con", np.max(con(xt)))
	
	return a, b

def rational_ratio_optimize(y, P, Q, a0, b0, norm = 2, **kwargs):
	r""" Find a locally optimal rational approximation in ratio form



	Parameters
	----------
	y
	P
	Q
	a0
	b0
	norm: [1, 2, np.inf]
		Norm in which to construct the approximation
	**kwargs
		Additional arguments to be passed to the optimization solver
	"""
	isreal = all([np.all(np.isreal(x)) for x in [y, P, Q, a0, b0]])

	m = P.shape[1]
	n = Q.shape[1]	
	x0 = np.hstack([a0, b0])


	if not isreal:
		x0 = x0.astype(np.complex)
		P = P.astype(np.complex)
		Q = Q.astype(np.complex)
		y = y.astype(np.complex)	
	
	if isreal and norm == 2:
		return _rational_ratio_optimize_pin_2norm_real(y, P, Q, a0, b0, **kwargs)
			
	elif (not isreal) and norm == 2:
		return _rational_ratio_optimize_pin_2norm_complex(y, P, Q, a0, b0, **kwargs)

	else:
		mess = f"The combination of norm={norm} and "
		if isreal:
			mess +="real data "
		else:
			mess +="complex data "
		mess+= "is not implemented"
		raise NotImplementedError(mess)
 
		





import numpy as np
import scipy.linalg
from iterprinter import IterationPrinter
from .rational import RationalBarycentric

def eval_barcentric(xeval, x, y, I, a, b):
	"""
	Parameters
	----------
	xeval: np.array
		Locations to evaluate the function
	x: np.array
		input coordinates for fit
	"""

	with np.errstate(divide='ignore',invalid='ignore'):
		A = np.hstack([ 1. /(xeval - xi) for xi in zip(x[I])])
		# For the rows that have a divide by zero error, 
		# we replace with value corresponding to the limit
		for row in np.argwhere(~np.all(np.isfinite(A), axis = 1)):
			A[row] = np.zeros(A.shape[1])
			A[row, np.argmin(np.abs(xeval[row] - x[I])).flatten()] = 1 
		
	num = np.einsum('ij,j...->i...', A, a)
	denom = np.einsum('ij,j->i', A,b)
	reval = np.einsum('i...,i->i...', num, 1./denom)
	
	return reval	



def eval_aaa(xeval, x, y, I, b):
	"""
	Parameters
	----------
	xeval: np.array
		Locations to evaluate the function
	x: np.array
		input coordinates for fit
	"""
	xeval = xeval.reshape(-1,1)
	with np.errstate(divide='ignore',invalid='ignore'):
		A = np.hstack([ bi /(xeval - xi) for bi, xi in zip(b, x[I])])
		# For the rows that have a divide by zero error, 
		# we replace with value corresponding to the limit
		for row in np.argwhere(~np.all(np.isfinite(A), axis = 1)):
			A[row] = np.zeros(A.shape[1])
			A[row, np.argmin(np.abs(xeval[row] - x[I])).flatten()] = 1 
		
	num = np.einsum('ij,j...->i...', A, y[I])
	denom = np.sum(A, axis = 1)
	reval = np.einsum('i...,i->i...', num, 1./denom)
	
	return reval	




def _build_cauchy(x,y):
	return 1./(np.tile(x.reshape(-1,1), (1,len(y))) - np.tile(y.reshape(1,-1), (len(x),1)))


def aaa(x, y, degree = None, tol = None, verbose = True):
	r""" A vector-valued Adaptive Anderson-Antoulas implementation

	Parameters
	----------
	x: np.array (M,)
		input coordinates
	y: np.array (M,...)
		output coordinates
	"""

	assert not ((degree is None) and (tol is None)), "One or both of 'degree' and 'tol' must be specified" 

	if degree is None:
		degree = len(x)//2 - 1


	mismatch = y
	I = np.zeros(len(y), dtype = np.bool)	# Index of x values used as barycentric nodes 

	if verbose:
		printer = IterationPrinter(it = '4d', res = '20.16e', cond = '8.3e')
		printer.print_header(it = 'iter', res = 'Frobenius norm residual', cond = 'condition number')

	for it in range(degree+1):
		# TODO: Is sum best? or how about maximum?
		residual = np.sum(np.abs(mismatch), axis = tuple(range(1,len(y[0].shape)+1)))
		residual[I] = 0 	# zero out residual at nodes that have already been selected
		
		Inew = np.argmax(residual)
		I[Inew] = True

		# Construct the Cauchy matrix 
		C = _build_cauchy(x[~I], x[I])
	 
		# Build the Loewner matrix associated with each input
		Lten = []
		for idx in np.ndindex(y[0].shape):
			Lten.append( (C.T * y[(~I,*idx)]).T - C*y[(I,*idx)] )
		L = np.vstack(Lten)
			
		# Compute coefficients for denominator polynomial
		U, s, VH = scipy.linalg.svd(L, full_matrices = False, compute_uv = True, overwrite_a = True)
		b = VH.conj().T[:,-1] 
		if len(s) >= 2:
			with np.errstate(divide='ignore',invalid='ignore'):
				cond = s[0]*np.sqrt(2)/(s[-2] - s[-1])
		else:
			cond = None

		mismatch = y - eval_aaa(x, x, y, I, b) 	
		res_norm = np.sqrt(np.sum(np.abs(mismatch)**2))
	
		if verbose:
			printer.print_iter(it=it, res = res_norm, cond = cond)	

		if tol is not None:
			if tol > res_norm:
				if verbose: print("terminated due to small residual")
				break

	return I, b 	

class AAARationalApproximation(RationalBarycentric):
	
	def __init__(self, degree = None, tol = None, verbose = True):
		self.degree = degree
		self.tol = tol
		self.verbose = verbose

	def fit(self, X, y):
		X = np.array(X)
		self.y = np.array(y)
		assert len(X) == len(y), "Length of X and y do not match"
		self.x = X.flatten()
		assert len(self.x) == len(y), "AAA only supports scalar-valued inputs"

		self.I, self.b = aaa(self.x, self.y, degree = self.degree, tol = self.tol, verbose = self.verbose)

	def __call__(self, X):
		x = np.array(X).flatten().reshape(-1,1)
		assert len(x) == len(X), "X must be a scalar-valued input"
		return eval_aaa(x, self.x, self.y, self.I, self.b)
		

class AAALawsonRationalApproximation(RationalBarycentric):
	pass






import numpy as np
from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment


def hungarian_sort(a, b):
	r""" Plug in replacement for marriage_sort that uses Hungarian algorithm for an optimal pairing.

	"""
	a = np.array(a).flatten().reshape(-1,1)
	b = np.array(b).flatten().reshape(-1,1)
	assert a.shape == b.shape, "a and b must be the same shape"
	X = cdist(np.hstack([a.real, a.imag]), np.hstack([b.real, b.imag]))
	row, col = linear_sum_assignment(X)
	I = np.argsort(row)
	return col[I]	

def sorted_norm(a, b, ord=2):
	r""" Compute the norm of the optimaly sorted vectors

	"""
	I = hungarian_sort(a, b)	
	return np.linalg.norm(a - b[I], ord)



r""" Basic polynomial classes

These classes use the polynomial representations included in numpy

"""

import numpy as np
from numpy.polynomial.polynomial import polyvander, polyder, polyroots
from numpy.polynomial.legendre import legvander, legder, legroots 
from numpy.polynomial.chebyshev import chebvander, chebder, chebroots
from numpy.polynomial.hermite import hermvander, hermder, hermroots
from numpy.polynomial.laguerre import lagvander, lagder, lagroots
from functools import lru_cache
from .index import *

class PolynomialBasis:

	def basis(self):
		r""" Alias for vandermonde(X) where X is the points where the basis was initialized

		Returns
		-------
		V: np.array (M, N)
			generalized Vandermonde matrix evaluated using the points the basis was initialized with
		"""
		raise NotImplementedError
	
		
	def vandermonde(self, X):
		r""" Construct the generalized Vandermonde matrix associated with the polynomial basis 

		Parameters
		----------
		X: np.array (M, dim)
			Coordinates at which to evaluate the basis at 

		Returns
		-------
		V: np.array (M, N)
			generalized Vandermonde matrix 
		"""
		raise NotImplementedError

	def vandermonde_derivative(self, X):
		raise NotImplementedError


class TensorProductPolynomialBasis(PolynomialBasis):
	def _vander(self, *args, **kwargs):
		raise NotImplementedError
	def _der(self, *args, **kwargs):
		raise NotImplementedError
	def roots(self, *args, **kwargs):
		raise NotImplementedError

	def __init__(self, X, degree):
		try:
			self.degree = int(degree)
			self._indices = total_degree_index(X.shape[1], degree)
			self.mode = 'total'
		except (TypeError, ValueError):
			self.degree = np.copy(degree)
			self._indices = max_degree_index(self.degree)
			self.mode = 'max'
	
		self._set_scale(X)
		self.X = np.copy(X)
		self.dim = X.shape[1]

	def _set_scale(self, X):	
		self._lb = np.min(X, axis = 0)
		self._ub = np.max(X, axis = 0)

	def _scale(self, X):
		r""" Scale coordinates to [-1,1]
		"""
		return 2*(X-self._lb[None,:])/(self._ub[None,:] - self._lb[None,:]) - 1

	def _inv_scale(self, X):
		return X*(self._ub[None,:] - self._lb[None,:])/2.0 + (self._ub[None,:] + self._lb[None,:])/2.0

	def vandermonde(self, X):
		X = self._scale(X)
		
		if self.mode == 'total':		 
			V_coordinate = [self._vander(X[:,k], self.degree) for k in range(self.dim)]
		elif self.mode == 'max':
			V_coordinate = [self._vander(X[:,k], d) for k,d in enumerate(self.degree)]
			
		
		V = np.ones((X.shape[0], len(self._indices)), dtype = X.dtype)
		for j, alpha in enumerate(self._indices):
			for k in range(self.dim):
				V[:,j] *= V_coordinate[k][:,alpha[k]]
		return V
	
	def vandermonde_derivative(self, X):
		r""" Construct a column-wise derivative of the generalized Vandermonde matrix
		"""
		M, dim = X.shape
		N = len(self._indices)
		DV = np.ones((M, N, dim), dtype = X.dtype)

		Y = self._scale(X) 

		if self.mode == 'total':		 
			V_coordinate = [self._vander(Y[:,k], self.degree) for k in range(self.dim)]
		elif self.mode == 'max':
			V_coordinate = [self._vander(Y[:,k], d) for k,d in enumerate(self.degree)]

		for k in range(dim):
			for j, alpha in enumerate(self._indices):
				for q in range(self.dim):
					if q == k:
						if self.mode == 'total':
							DV[:,j,k] *= V_coordinate[q][:,0:-1] @ self._Dmat[alpha[q],:] 
						elif self.mode == 'max':
							DV[:,j,k] *= V_coordinate[q][:,0:-1] @ self._Dmat[alpha[q],:self.degree[q]]
					else:
						DV[:,j,k] *= V_coordinate[q][:,alpha[q]]

			DV[:,:,k] *= self._dscale[k] 			

		return DV	
		  
	

	@lru_cache(maxsize = 1)
	def basis(self):
		r""" The basis for the input coordinates
		""" 
		return self.vandermonde(self.X)
	
	@property
	@lru_cache(maxsize = 1)
	def _Dmat(self):
		r""" The matrix specifying the action of the derivative operator in this basis
		"""
		if self.mode == 'total':
			max_degree = self.degree
		elif self.mode == 'max':
			max_degree = max(self.degree)
		Dmat = np.zeros( (max_degree+1, max_degree))
		I = np.eye(max_degree + 1)
		for j in range(max_degree + 1):
			Dmat[j,:] = self._der(I[:,j])
		return Dmat


	@property
	@lru_cache(maxsize = 1)
	def _dscale(self):
		r""" Derivative of the scaling applied to the coordinates
		"""
		# As we assume the transformation is linear, we simply compute the finite difference
		# with a unit step size
		XX = np.zeros((2, self.dim))
		XX[1,:] = 1
		sXX = self._scale(XX)
		dscale = sXX[1] - sXX[0]
		return dscale

class MonomialPolynomialBasis(TensorProductPolynomialBasis):
	def _vander(self, *args, **kwargs):
		return polyvander(*args, **kwargs)
	def _der(self, *args, **kwargs):
		return polyder(*args, **kwargs)
	def roots(self, *args, **kwargs):
		return self._inv_scale(polyroots(*args, **kwargs))


class LegendrePolynomialBasis(TensorProductPolynomialBasis):
	def _vander(self, *args, **kwargs):
		return legvander(*args, **kwargs)
	def _der(self, *args, **kwargs):
		return legder(*args, **kwargs)
	def roots(self, *args, **kwargs):
		return self._inv_scale(legroots(*args, **kwargs))

class ChebyshevPolynomialBasis(TensorProductPolynomialBasis):
	def _vander(self, *args, **kwargs):
		return chebvander(*args, **kwargs)
	def _der(self, *args, **kwargs):
		return chebder(*args, **kwargs)
	def roots(self, *args, **kwargs):
		return self._inv_scale(chebroots(*args, **kwargs))

class HermitePolynomialBasis(TensorProductPolynomialBasis):
	def _vander(self, *args, **kwargs):
		return hermvander(*args, **kwargs)
	def _der(self, *args, **kwargs):
		return hermder(*args, **kwargs)
	def roots(self, *args, **kwargs):
		return self._inv_scale(hermroots(*args, **kwargs))
	
	def _set_scale(self, X):
		self._mean = np.mean(X, axis = 0)
		self._std = np.std(X, axis = 0)
	
	def _scale(self, X):		
		return (X - self._mean[None,:])/self._std[None,:]/np.sqrt(2)

	def _inv_scale(self, X):
		return np.sqrt(2)*self._std[None,:]*X + self._mean[None,:]

class LaguerrePolynomialBasis(TensorProductPolynomialBasis):
	def _vander(self, *args, **kwargs):
		return lagvander(*args, **kwargs)
	def _der(self, *args, **kwargs):
		return lagder(*args, **kwargs)
	def roots(self, *args, **kwargs):
		return self._inv_scale(lagroots(*args, **kwargs))

	def _set_scale(self, X):
		r""" Laguerre polynomial expects x[i] to be distributed like exp[-lam*x] on [0,infty)
		so we shift so that all entries are positive and then set a scaling
		"""
		self._lb = np.min(X, axis = 0)
		self._a = 1./np.mean(X - self._lb[None,:], axis = 0)
		
	def _scale(self, X):
		return self._a[None,:]*(X - self._lb[None,:])
		
	def _inv_scale(self, X):
		return X/self._a[None,:] + self._lb[None,:]





	



import os
from setuptools import setup


install_requires = [
	'numpy',
	'scipy',
	'iterprinter',
	'cvxopt',
	] 

# CVXPY is licensed under the GPL
# My understanding is that this implies the _tests_ are under a GPL license
# but the underlying code in polyrat is still able to be under a more permissive license
# See discussion: https://opensource.stackexchange.com/a/7510
test_requires = [
	'pytest',
	'cvxpy',
]


with open('README.md', 'r') as f:
	long_description = f.read()

ns = {}
with open('polyrat/version.py') as f:
	exec(f.read(), ns)

version = ns['__version__']

setup(name='polyrat',
	version = version,
	url = 'https://github.com/jeffrey-hokanson/PolyRat',
	description = 'Polynomial and rational function library',
	long_description = long_description,
	long_description_content_type = 'text/markdown', 
	author = 'Jeffrey M. Hokanson',
	author_email = 'jeffrey@hokanson.us',
	packages = ['polyrat',],
	install_requires = install_requires,
	test_requires = test_requires,
	python_requires='>=3.6',
	classifiers = [
		'Development Status :: 4 - Beta',
		"Programming Language :: Python :: 3",
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',	
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Mathematics'
	]
	)



import numpy as np
from .basis import *
from copy import deepcopy
import scipy.linalg

class Polynomial:
	def __init__(self, basis, coef):
		self.basis = deepcopy(basis)
		self.coef = np.copy(coef)

	def __call__(self, X):
		return self.basis.vandermonde(X) @ self.coef

	def eval(self, X):
		return self.basis.vandermonde(X) @ self.coef

	def roots(self, *args, **kwargs): 
		return self.basis.roots(self.coef, *args, **kwargs)	


def _polynomial_fit_least_squares(P, y):
	coef, _, _, _ = scipy.linalg.lstsq(P, y)
	return coef.flatten()	

class PolynomialApproximation(Polynomial):
	def __init__(self, degree, Basis = None, norm = 2):
		if Basis is None:
			from .arnoldi import ArnoldiPolynomialBasis
			Basis = ArnoldiPolynomialBasis
		self.Basis = Basis
		self.degree = degree
		self.norm = norm

	def fit(self, X, y):
		self.basis = self.Basis(X, self.degree)
		P = self.basis.basis()
		if self.norm == 2:
			self.coef = _polynomial_fit_least_squares(P, y)
		else:
			raise NotImplementedError
		





""" Routines implementing variants of the Sanathanan-Koerner iteration
"""
import numpy as np
from numpy.linalg import LinAlgError
import scipy.linalg
import scipy.optimize
from .basis import *
from .arnoldi import *
from .polynomial import *
from iterprinter import IterationPrinter



def _minimize_2_norm(A):
	r"""
		Solve the optimization problem 
		
		min_x  || A @ x ||_2
	"""
	U, s, VH = scipy.linalg.svd(A, full_matrices = False)
		
	# Condition number of singular vectors, cf. Stewart 01: Eq. 3.16
	with np.errstate(divide = 'ignore'):
		cond = s[0]*np.sqrt(2)/(s[-2] - s[-1])
	return VH.T.conj()[:,-1], cond


def _minimize_inf_norm(A):
	raise NotImplementedError

def _minimize_1_norm(A):
	raise NotImplementedError


#def _minimize_1_norm_real(A):
#	r"""
#		Solve the optimization problem 
#		
#		min_x  || A @ x ||_1  s.t.  ||x||_2 = 1
#	"""
#	m, n = A.shape
#
#	U, s, VH = np.linalg.svd(A, full_matrices = False)
#	print(m,n, *U.shape)
#	A_ub = np.vstack([
#			# A x - t <= 0
#			np.hstack([A, -np.eye(m)]),
#			# -A x - t <= 0 
#			np.hstack([-A, -np.eye(m)])
#			])
#	b_ub = np.zeros(2*m)
#
#	# Pin one of the variables so we have a non-zero solution
#	A_eq = np.zeros((1, m +n))
#	A_eq[0,0] = 1
#	b_eq = np.ones((1,))
#
#	# Objective: minimize the sum of the upper bounds
#	c = np.zeros(n + m)
#	c[n:] = 1.
#
#	lb = -np.inf*np.ones(m+n)
#	ub = np.inf*np.ones(m+n)
#	lb[n:] = 0
#	
#	bounds = [[None, None] for i in range(m+n)]
#	for i in range(m): bounds[n+i][0] = 0
#
#	res = scipy.optimize.linprog(c, A_ub = A_ub, b_ub = b_ub, A_eq = A_eq, b_eq = b_eq, bounds = bounds,
#			options = {'presolve': True, 'autoscale': True, 'lstsq': True})
#	
#	#y = res.x[:n]
#	# U @ y = A @ x 
#	# U @ y = U @ np.diag(s) @ VH @ x
#	# y = np.diag(s) @ VH @ x
#	#x = (VH.conj().T @ (y/s ))
#	x = res.x[:n]
#	x /= np.linalg.norm(x)
#	x *= np.sign(x[0])
#	return x, s
		

def linearized_ratfit(X, y, num_degree, denom_degree, Basis = ArnoldiPolynomialBasis, simultaneous = False):
	r""" This solves the linearized rational approximation problem

	See: AKL+19x
	"""
	num_basis = Basis(X, num_degree)
	denom_basis = Basis(X, denom_degree)
	P = num_basis.basis()
	Q = denom_basis.basis()		

	# diag(y) @ Q
	yQ = np.multiply(y[:,None], Q)
	

	if simultaneous:
		A = np.hstack([P, -yQ])
		x, cond = _minimize_2_norm(A)
		a = x[:P.shape[1]]
		b = x[-Q.shape[1]:]

	elif Basis == ArnoldiPolynomialBasis:
		# In AKL+19x implementation they reduce to a problem only over b
		# by using the pseudoinverse to implicitly solve for a
		# (much like in Variable Projection)
		# In this case the basis P has orthonormal columns, so we have no
		# need for the pseudoinverse
		W = P @ (P.conj().T @ yQ) - yQ
		b, cond = _minimize_2_norm(W)
		
		# and then idenify a via the pseudo-inverse
		a = P.conj().T @ (yQ @ b)
	else:
		W = P @ np.linalg.lstsq(P, yQ, rcond = None)[0] - yQ
		b, cond = _minimize_2_norm(W)
		a = np.linalg.lstsq(P, yQ @ b, rcond = None)[0]

			
	numerator = Polynomial(num_basis, a)
	denominator = Polynomial(denom_basis, b)

	return numerator, denominator

def skfit(y, P, Q, maxiter = 20, verbose = True, history = False, denom0 = None, norm = 2, xtol = 1e-7):
	r"""


	Returns
	-------
	a: np.array 
		Numerator coefficients
	b: np.array
		Denominator coefficients
	"""	
	if denom0 is None:
		denom = np.ones(len(y), dtype = y.dtype)
	else:
		assert len(denom0) == len(y)
		denom = np.array(denom0)

	if verbose:
		printer = IterationPrinter(it = '4d', res_norm = '21.15e', delta_fit = '8.2e', cond = '8.2e')
		printer.print_header(it = 'iter', res_norm = 'residual norm', delta_fit = 'delta fit', cond = 'cond')
	
	if history:
		hist = []

	# As there are no guarntees about convergence,
	# we record the best iteration
	best_res_norm = np.inf
	best_sol = None	
	
	# For comparison with current iterate to determine termination
	fit_old = np.zeros(y.shape[0], dtype = y.dtype)

	
	for it in range(maxiter):
		A = np.hstack([ 
				np.multiply((1./denom)[:,None], P), 	
				np.multiply((-y/denom)[:,None], Q)
			])
		x, cond = _minimize_2_norm(A)
		
		a = x[:P.shape[1]]
		b = x[-Q.shape[1]:]
		Pa = P @ a
		Qb = Q @ b
		fit = Pa/Qb
		res_norm = np.linalg.norm(y - fit, norm)

		if res_norm < best_res_norm:
			best_res_norm = res_norm
			best_sol = [a, b]

		delta_fit = np.linalg.norm(fit - fit_old, norm)		
		
		if history:
			hist.append({'fit': fit, 'cond': cond})

		if verbose:
			printer.print_iter(it = it, delta_fit = delta_fit, res_norm = res_norm, cond = cond) 

		if delta_fit < xtol:
			break

		# Copy over data for next loop
		denom = Qb 
		fit_old = fit

	if history:	
		return best_sol + [hist]
	else:
		return best_sol



def skfit_rebase(X, y, num_degree, denom_degree, maxiter = 20, verbose = True, 
	xtol = 1e-7, history = False, denom0 = None, norm = 2):
	r""" The SK-iteration, but at each step use Vandermonde with Arnoldi to construct a new basis

	Parameters
	----------
	X: np.array (M,dim)
		
	y: np.array (M,)


	Returns
	-------
			
	"""

	if history:
		hist = []


	if denom0 is None:
		denom = np.ones(X.shape[0], dtype = X.dtype)
	else:
		assert denom0.shape[0] == X.shape[0]
		denom = denom0


	if np.isclose(norm, 2):
		linearized_solution = _minimize_2_norm
	elif ~np.isfinite(norm):
		linearized_solution = _minimize_inf_norm
	else: 
		raise NotImplementedError
	
	if verbose:
		printer = IterationPrinter(it = '4d', res_norm = '21.15e', delta_fit = '8.2e', cond = '8.2e')
		printer.print_header(it = 'iter', res_norm = 'residual norm', delta_fit = 'delta fit', cond = 'cond')

	# As there are no guarntees about convergence,
	# we record the best iteration
	best_res_norm = np.inf
	best_sol = None	
	
	# For comparison with current iterate to determine termination
	fit_old = np.zeros(y.shape[0], dtype = X.dtype)

	for it in range(maxiter):
		try:
			num_basis = ArnoldiPolynomialBasis(X, num_degree, weight = 1./denom)
			denom_basis = ArnoldiPolynomialBasis(X, denom_degree, weight = 1./denom)
			P = num_basis.basis()
			Q = denom_basis.basis()	
			#P, RP, _ = vandermonde_arnoldi_CGS(X, num_degree, weight = 1./denom)	
			#Q, RQ, _ = vandermonde_arnoldi_CGS(X, denom_degree, weight = 1./denom)	
			
			A = np.hstack([P, np.multiply(-y[:,None], Q) ])
			x, cond = linearized_solution(A)
			a = x[:P.shape[1]]
			b = x[-Q.shape[1]:]
			
			Pa = P @ a
			Qb = Q @ b

			fit = Pa/Qb

			delta_fit = np.linalg.norm(fit - fit_old, norm)		
			res_norm = np.linalg.norm(fit - y, norm)
		
		except (LinAlgError, ValueError) as e:
			if verbose: print(e)
			break
	

		# If we have improved the fit, append this 
		if res_norm < best_res_norm:
			numerator = Polynomial(num_basis, a)
			denominator = Polynomial(denom_basis, b)
			best_sol = [numerator, denominator]
			best_res_norm = res_norm

		if history:
			hist.append({'fit': fit, 'cond':cond})

		if verbose:
			printer.print_iter(it = it, delta_fit = delta_fit, res_norm = res_norm, cond = cond) 

		if delta_fit < xtol:
			break

		denom = np.abs(denom * Qb)
		denom[denom == 0.] = 1.
		fit_old = fit

	if history:	
		return best_sol + [hist]
	else:
		return best_sol
